import logging
import sys
from re import sub
from pyscilog.filter import LogFilter
from pyscilog.writer import Writer
from pyscilog.state import State

state = State()
log_filter = LogFilter()
fmt = "%(asctime)s - %(shortname)-18.18s %(subprocess)s%(memory)s%(separator)s%(message)s"
datefmt = '%H:%M:%S'  # '%H:%M:%S.%f'

# this will be a null handler
null_handler = logging.NullHandler()


class ColorStrippingFormatter(logging.Formatter):
    def __init__(self, fmt, datefmt, strip=True):
        logging.Formatter.__init__(self, fmt, datefmt)
        self.strip = strip

    def label(self, record):
        if record.levelno < logging.WARNING:
            return "INFO", "37;42"
        elif record.levelno < logging.ERROR:
            return "WARNING", "37;43"
        elif record.levelno < logging.CRITICAL:
            return "ERROR", "5;41"
        else:
            return "CRITICAL", "5;41"

    def format(self, record):
        """Uses parent class to format record, then strips off colors"""
        msg = logging.Formatter.format(self, record)
        label, color = self.label(record)
        if self.strip:
            return "{:10s}{}".format(label, sub("\033\\[[0-9]+m", "", msg, 0))
        else:
            return "\033[1;{}m{:10s}\033[0m{}".format(color, label, msg)


_logfile_formatter = ColorStrippingFormatter(fmt, datefmt, strip=True)
_console_formatter = ColorStrippingFormatter(fmt, datefmt, strip=False)


class LoggerWrapper:
    def __init__(self, logger, verbose=None, log_verbose=None):
        self.logger = logger
        logger.propagate = False

        self.console_handler = logging.StreamHandler(sys.stderr)
        self.console_handler.setFormatter(_console_formatter)

        self.logfile_handler = logging.handlers.MemoryHandler(1, logging.DEBUG, state['file_handler'] or null_handler)
        self.logfile_handler.setFormatter(_logfile_formatter)

        # set verbosity levels
        self._verbose = self._log_verbose = None
        self.verbosity(verbose if verbose is not None else state['verbosity'])
        self.log_verbosity(log_verbose if log_verbose is not None else state['log_verbosity'])

        # other init
        self.logger.addHandler(self.console_handler)
        self.logger.addHandler(self.logfile_handler)
        self.logger.addFilter(log_filter)

    def verbosity(self, set_verb=None):
        if set_verb is not None:
            self._verbose = set_verb
            self.console_handler.setLevel(logging.INFO - set_verb)
            if self._log_verbose is None:
                self.logfile_handler.setLevel(logging.INFO - set_verb)
        return self._verbose

    def log_verbosity(self, set_verb=None):
        if set_verb is not None:
            self._log_verbose = set_verb
            self.logfile_handler.setLevel(logging.INFO - set_verb)
        return self._log_verbose if self._log_verbose is not None else self._verbose

    def __call__(self, level, color=None):
        """
        Function call operator on logger. Use to issue messages at different verbosity levels.
        E.g. log(2).print("message" will issue a message at level logging.INFO - 2.)
        An optional color argument will colorize the message.
        Returns:
            A writer object (to which a message may be sent with "<<")
        """
        # effective verbosity level is either set explicitly when the writer is created, or else use global level
        return Writer(self.logger, logging.INFO - level, color=color)

    def warn(self, msg, color=None, print_once=None):
        """
        Wrapper for log.warn
        """
        Writer(self.logger, logging.WARN, color=color).write(msg, print_once=print_once)

    warning = warn

    def error(self, msg, color="red", print_once=None):
        """
        Wrapper for log.error
        """
        Writer(self.logger, logging.ERROR, color=color).write(msg, print_once=print_once)

    def info(self, msg, color=None, print_once=None):
        """
        Wrapper for log.info
        """
        Writer(self.logger, logging.INFO, color=color).write(msg, print_once=print_once)

    def critical(self, msg, color=None, print_once=None):
        """
        Wrapper for log.critical
        """
        Writer(self.logger, logging.CRITICAL, color=color).write(msg, print_once=print_once)

    def debug(self, msg, color=None, print_once=None):
        """
        Wrapper for log.debug
        """
        Writer(self.logger, logging.DEBUG, color=color).write(msg, print_once=print_once)

    def exception(self, msg, color=None, print_once=None):
        """
        Wrapper for log.exception
        """
        Writer(self.logger, logging.EXCEPTION, color=color).write(msg, print_once=print_once)

    def print(self, *args):
        return self.info(" ".join(map(str, args)))

    def write(self, message, level=logging.INFO, verbosity=0, print_once=None, color=None):
        # apply verbosity only to INFO levels
        if level == logging.INFO:
            level -= int(verbosity)
        Writer(self.logger, level, color=color).write(message, print_once=print_once)


def log_to_file(filename, append=False):
    if not state['file_handler']:
        _file_handler = logging.FileHandler(filename, mode='a' if append else 'w')
        _file_handler.setLevel(logging.DEBUG)
        _file_handler.setFormatter(_logfile_formatter)
        # set it as the target for the existing wrappers' handlers
        for wrapper in state['loggers'].values():
            wrapper.logfile_handler.setTarget(_file_handler)


def set_boring(boring=True):
    _console_formatter.strip = boring



from dataclasses import dataclass
from pyscilog.state import State

SEPERATOR = "================================{}=================================="


@dataclass
class ColorMap:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    NOBOLD = '\033[0m'


color_dict = dict(red=ColorMap.FAIL, green=ColorMap.OKGREEN, yellow=ColorMap.WARNING, blue=ColorMap.OKBLUE, white="")

state = State()


def cprint(string: str, col="red", bold=True):
    if state['silent']:
        return string

    ss = color_dict.get(col)
    if ss is None:
        raise ValueError("unknown color '{}'".format(col))

    ss = "%s%s%s" % (ss, string, ColorMap.ENDC)
    if bold:
        ss = "%s%s%s" % (ColorMap.BOLD, ss, ColorMap.NOBOLD)

    return ss


def disable_colors():
    state.silent = True


def separate(strin=None, D=1):
    if D != 1:
        return cprint(SEPERATOR.format("=" * len(strin)))
    else:
        return cprint(SEPERATOR.format(strin))


def title(strin, big=False):
    print()
    print()
    if big:
        print(separate(strin, D=0))

    print(separate(strin))

    if big:
        print(separate(strin, D=0))

    print()


def disable():
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''



from pyscilog import cprint
from typing import Set, Any


class Writer:
    """A default writer logs messages to a logger"""
    __print_once_keys: Set[Any] = set()

    def __init__(self, logger, level, color=None, bold=None):
        self.logger = logger
        self.level = level
        self.color = (color or "red") if bold else color
        self.bold = bool(color) if bold is None else bold

    def write(self, message, level_override=None, print_once: Any = None):
        if print_once is not None:
            if print_once in Writer.__print_once_keys:
                return
            Writer.__print_once_keys = set(Writer.__print_once_keys.union(set(print_once)))

        message = message.rstrip()
        if self.color and message:  # do not colorize empty messages, else "\n" is issued independently
            message = cprint(message, col=self.color, bold=self.bold)
        self.logger.log(self.level if level_override is None else level_override, message)

    def print(self, *args):
        return self.write(" ".join(map(str, args)))



from __future__ import print_function
import logging
import logging.handlers

from pyscilog.color import cprint
from pyscilog.filter import LogFilter, get_subprocess_label, set_subprocess_label
from pyscilog.handlers import _sigusr1_handler, _sigusr2_handler
from pyscilog.wrapper import LoggerWrapper, datefmt, set_boring, log_to_file
from pyscilog.state import State
from pyscilog.handlers import init_handlers

state = State()
init_handlers()


def get_log_filename():
    """
    Returns log filename if log_to_file has been called previously, None otherwise
    """
    if not state['file_handler']:
        return None
    return state['file_handler'].baseFilename


def enable_memory_logging(level=1):
    LogFilter.setMemoryLogging((level or 0) % 3)  # level is 0/1/2


def init(app_name):
    if state['root_logger'] is None:
        logging.basicConfig(datefmt=datefmt)
        state['app_name'] = app_name
        state['root_logger'] = logging.getLogger(app_name)
        state['root_logger'].setLevel(logging.DEBUG)
        state['log'] = state['loggers'][''] = LoggerWrapper(state['root_logger'])


def get_logger(name, verbose=None, log_verbose=None):
    """Creates a new logger (or returns one, if already created)"""
    init("app")
    if name in state['loggers']:
        return state['loggers'][name]

    logger = logging.getLogger("{}.{}".format(state['app_name'], name))
    lw = state['loggers'][name] = LoggerWrapper(logger, verbose, log_verbose)
    lw(2).print("logger initialized")

    return lw


def set_silent(log_name):
    """Silences the specified sublogger(s)"""
    state['log'].print(cprint("set silent: %s" % log_name, col="red"))
    if isinstance(log_name, str):
        get_logger(log_name).logger.setLevel(logging.CRITICAL)
    elif type(log_name) is list:
        for name in log_name:
            get_logger(name).logger.setLevel(logging.CRITICAL)


def set_loud(log_name):
    """Un-silences the specified sublogger(s)"""
    state['log'].print(cprint("set loud: %s" % log_name, col="green"))
    if isinstance(log_name, str):
        get_logger(log_name).logger.setLevel(logging.DEBUG)
    elif type(log_name) is list:
        for name in log_name:
            get_logger(name).logger.setLevel(logging.DEBUG)



import re
from typing import Union, List
from pyscilog import get_logger
from pyscilog.state import State

state = State()


def set_verbosity(verbosity: Union[int, List[int], str]):
    if verbosity is None:
        state['verbosity'] = 0
        return
    # ensure verbosity is turned into a list.
    if type(verbosity) is int:
        verbosity = [verbosity]
    elif isinstance(verbosity, str):
        verbosity = verbosity.split(",")
    elif not isinstance(verbosity, (list, tuple)):
        raise TypeError("can't parse verbosity specification of type '{}'".format(type(verbosity)))
    for element in verbosity:
        if type(element) is int or re.match("^[0-9]+$", element):
            state['verbosity'] = int(element)
            state['log'](0, "green").print("set global console verbosity level {}".format(state['verbosity']))
        else:
            m = re.match("^(.+)=([0-9]+)$", element)
            if not m:
                raise ValueError("can't parse verbosity specification '{}'".format(element))
            logger = get_logger(m.group(1))
            level = int(m.group(2))
            logger.verbosity(level)
            logger(0, "green").print("set console verbosity level {}={}".format(m.group(1), level))


def get_verbosity(verbosity: Union[List[int], int, str]):
    if verbosity is None:
        state['log_verbosity'] = None  # None means follow console default
        return
    # ensure verbosity is turned into a list.
    if type(verbosity) is int:
        verbosity = [verbosity]
    elif isinstance(verbosity, str):
        verbosity = verbosity.split(",")
    elif not isinstance(verbosity, (list, tuple)):
        raise TypeError("can't parse verbosity specification of type '{}'".format(type(verbosity)))
    for element in verbosity:
        if type(element) is int or re.match("^[0-9]+$", element):
            state['log_verbosity'] = int(element)
            if state['log_verbosity'] is not None:
                state['log'](0, "green").print("set global log verbosity level {}".format(state['log_verbosity']))
        else:
            m = re.match("^(.+)=([0-9]+)$", element)
            if not m:
                raise ValueError("can't parse verbosity specification '{}'".format(element))
            logger = get_logger(m.group(1))
            level = int(m.group(2))
            logger.log_verbosity(level)
            logger(0, "green").print("set log verbosity level {}={}".format(m.group(1), level))



import os
import psutil


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return getinstance


@singleton
class State(dict):
    pass


state = State()
state['silent'] = False

# this will be the handler for the log file
state['file_handler'] = None

# dict of logger wrappers created by the application
state['loggers'] = {}

state['root_logger'] = None
state['log'] = None

# global verbosity levels (used for loggers for which an explicit level is not set)
state['verbosity'] = 0
state['log_verbosity'] = None

state['subprocess_label'] = None
state['parent_process'] = psutil.Process(os.getpid())
state['log_memory_totals'] = True
state['log_memory_types'] = None, "rss vms".split(), "rss pss".split()



import os
import signal

from pyscilog import LogFilter


def _sigusr1_handler(signum, frame):
    level = 2 if LogFilter._log_memory == 1 else 1
    print("pid {} received USR1: memory logging level {}".format(os.getpid(), level))
    LogFilter.setMemoryLogging(level)


def _sigusr2_handler(signum, frame):
    print("pid {} received USR2: disabling memory logging".format(os.getpid()))
    LogFilter.setMemoryLogging(0)


def init_handlers():
    signal.signal(signal.SIGUSR1, _sigusr1_handler)
    signal.signal(signal.SIGUSR2, _sigusr2_handler)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

tests_require = ['pytest', 'mypy', 'pycodestyle', 'data-science-types']

extras_require = {
    'test': tests_require,
    'doc': ['sphinx', 'sphinx_rtd_theme'],
}

packages = find_packages(exclude=['tests'])


setup(
    name="pyscilog",
    version="0.1.2",
    author="Gijs Molenaar",
    author_email="gijs@pythonic.nl",
    description="the Python scientific logger",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gijzelaerr/pyscilog/",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires='>=3.6',
    extras_require=extras_require,
    install_requires=['psutil',
                      'dataclasses'],
    tests_require=tests_require,
    test_suite="tests",
)



import os
import time
from logging import Filter
import psutil
from multiprocessing import current_process
from pyscilog.color import cprint
from pyscilog.state import State

state = State()

GB = float(1024 ** 3)


def set_subprocess_label(label):
    """
    Sets the subprocess label explicitly
    """
    state['subprocess_label'] = label


def get_subprocess_label():
    """
    Returns the subprocess ID. For the main process, this is empty. For subprocesses
    (started by multiprocessing), this is "Pn" by default, where n is the process number.
    """
    if state['subprocess_label'] is None:
        name = current_process().name
        if name == "MainProcess":
            state['subprocess_label'] = ""
        else:
            state['subprocess_label'] = name.replace("Process-", "P")
    return state['subprocess_label']


class LogFilter(Filter):
    _log_memory = 0

    @staticmethod
    def setMemoryLogging(level):
        LogFilter._log_memory = level
        LogFilter._mem_ts = LogFilter._mem_totals_ts = LogFilter._children_ts = 0

    _mem = None
    _mem_totals = None
    _children = None
    _mem_ts = 0
    _mem_totals_ts = 0
    _children_ts = 0
    _mem_update = (
        0, .5, 2)  # full memory updates are a little costly, so do them only after this many seconds has elapsed
    _mem_totals_update = (0, 1, 5)
    _children_update = 3  # children updates even more so, so do it even less frequently

    """LogFilter augments the event by a few new attributes used by our formatter"""

    def filter(self, event):
        if not event.getMessage().strip():
            return False
        # short logger name (without app_name in front of it)
        setattr(event, 'shortname', event.name.split('.', 1)[1] if '.' in event.name else event.name)
        setattr(event, 'separator', '| ')
        memlevel = LogFilter._log_memory  # signal handler can change that midway through, so use this value
        if memlevel:
            # memory usage info
            t = time.time()
            if t - self._mem_ts > self._mem_update[memlevel]:
                keys = state['log_memory_types'][memlevel]
                # get memory for this process
                mi0 = psutil.Process(os.getpid()).memory_full_info() if memlevel == 2 else \
                    psutil.Process(os.getpid()).memory_info()
                mem = {key: getattr(mi0, key) / GB for key in keys}
                shm = psutil.virtual_memory().shared / GB
                # get total memory
                if state['log_memory_totals']:
                    if t - self._mem_totals_ts > self._mem_totals_update[memlevel]:
                        if t - self._children_ts > self._children_update:
                            self._children = [state['parent_process']] + list(state['parent_process'].children(recursive=True))
                            self._children_ts = time.time()
                        if len(self._children) > 1:
                            mis = []
                            # scan over children, ignoring ones that may have disappeared
                            for p in self._children:
                                try:
                                    mis.append(p.memory_full_info()) if memlevel == 2 else mis.append(p.memory_info())
                                except psutil.NoSuchProcess:
                                    pass
                            self._mem_totals = {key: sum([getattr(mi, key) for mi in mis]) / GB for key in keys}
                        self._mem_totals_ts = time.time()
                # form up string
                if self._mem_totals is None:
                    smem = ["{:.1f}".format(mem.get(key, 0)) for key in keys]
                else:
                    smem = ["{:.1f}/{:.1f}".format(mem.get(key, 0), self._mem_totals.get(key, 0)) for key in keys]
                smem.append("{:.1f}Gb".format(shm))
                self._mem = " ".join(smem)
                self._mem_ts = time.time()
            setattr(event, "memory", "[{}{}] ".format("*" if memlevel == 2 else "", self._mem))
            setattr(event, 'separator', '')
        else:
            setattr(event, "memory", "")
        # subprocess info
        subprocess_id = get_subprocess_label()
        if subprocess_id:
            setattr(event, "subprocess", cprint("[%s] " % subprocess_id, col="blue"))
            setattr(event, 'separator', '')
        else:
            setattr(event, "subprocess", "")
        return True
